# This is a simple store of OpenID data
import heapq, time, calendar, openid2rp

class InMemoryStore:
    nonce_lifetime = 300 #s

    def __init__(self):
        # ongoing logins
        self.logins = {}
        # received reply nonces
        self.reply_nonces = {}
        # established provider associations
        self.associations = {}
        # heap of (time, dictionary, key) triples
        self.expirations = []

    def expire(self):
        now = time.time()
        while True:
            t, d, k = self.expirations[0]
            if t > now:
                break
            del d[k]
            heapq.heappop(heap)

    def start_login(self, key, value):
        now = time.time()
        self.logins[key] = value
        heapq.heappush(self.expirations, (now+3600, self.logins, key))

    def get_login(self, key):
        return self.logins.get(key)

    def end_login(self, key):
        del self.logins[key]
        for i, k in enumerate(self.expirations):
            if k[1] is self.logins and k[2] == key:
                del self.expirations[i]
                # deletion in the middle is not supported by heapq,
                # so re-heapify instead
                heapq.heapify(self.expirations)
                break

    def add_nonce(self, nonce):
        utc = calendar.timegm(openid2rp.parse_nonce(nonce).utctimetuple())
        self.reply_nonces[nonce] = None
        heapq.heappush(self.expirations, (utc+self.nonce_lifetime, self.reply_nonces, nonce))

    def has_nonce(self, nonce):
        return nonce in self.reply_nonces

    def add_association(self, key, expires, value):
        self.associations[key] = value
        heapq.heappush(self.expirations, (expires, self.associations, key))

    def get_association(self, key):
        return self.associations.get(key)



import cPickle, webob, time, random, calendar
import openid2rp, memstore

class Openid2Middleware(object):
    def __init__(self, app, store = None):
        self.app = app
        if not store:
            store = memstore.InMemoryStore()
        self.store = store
        # not yet used
        self.return_to = None

    def __call__(self, environ, start_response):
        request = webob.Request(environ)

        if 'openid_identifier' in request.params:
            return self.login(request, start_response)

        if 'openid_return' in request.params:
            return self.returned(request, start_response)

        return self.app(environ, start_response)

    def login(self, req, start_response):
        kind, claimed_id = openid2rp.normalize_uri(req.params['openid_identifier'])
        if kind == 'xri':
            # XRIs are not supported. Report this to the application as an error
            return self.error(req.environ, start_response, 'XRIs are not supported')
        res = openid2rp.discover(claimed_id)
        if not res:
            # not an OpenID
            req.environ['openid2rp.notice'] = 'discovery failed'
            return self.app(req.environ, start_response)
        services, op_endpoint, op_local = res

        assoc = self.store.get_association(claimed_id)
        if not assoc:
            now = int(time.time())
            # XXX error handling
            assoc = openid2rp.associate(services, op_endpoint)
            assoc_handle = assoc['assoc_handle']
            expires = now + int(assoc['expires_in'])
            mac_key = assoc['mac_key']
            saved_assoc = cPickle.dumps((assoc_handle, mac_key))
            self.store.add_association(claimed_id, expires, saved_assoc)
            self.store.add_association(assoc_handle, expires, saved_assoc)
        else:
            assoc_handle, mac_key = cPickle.loads(assoc)

        session = str(random.getrandbits(40))
        self.store.start_login(session, cPickle.dumps((claimed_id, assoc_handle)))

        if self.return_to:
            return_to = self.return_to
            if '?' not in return_to:
                return_to += '?openid_return='+session
            else:
                return_to += '&openid_return='+session
        else:
            return_to = req.path_url+'?openid_return='+session
        redirect = openid2rp.request_authentication(
            services, op_endpoint, assoc_handle, return_to, 
            claimed=claimed_id, op_local=op_local)

        start_response('303 Go to OpenID provider', [('Location', redirect)])
        return []

    def returned(self, req, start_response):
        session = req.params.get('openid_return')
        dump = self.store.get_login(session)
        if not dump:
            req.environ['openid2rp.error'] = 'login session not found'
            return self.app(req.environ, start_response)
        claimed_id, assoc_handle = cPickle.loads(dump)
        self.store.end_login(session)
        
        qs = req.environ.get('QUERY_STRING')
        # XXX error handling
        assoc = self.store.get_association(assoc_handle)
        assoc_handle, mac_key = cPickle.loads(assoc)
        try:
            signed = openid2rp.authenticate({'assoc_handle':assoc_handle, 'mac_key':mac_key}, qs)
        except Exception, e:
            return self.error(req.environ, start_response, str(e))

        # Check for replay attacks
        nonce = req.params['openid.response_nonce']
        utc = calendar.timegm(openid2rp.parse_nonce(nonce).utctimetuple())
        if time.time()+self.store.nonce_lifetime < utc or self.store.has_nonce(nonce):
            return self.error(req.environ, start_response, 'replay attack detected')
        self.store.add_nonce(nonce)

        req.environ['openid2rp.identifier'] = claimed_id
        namespaces = openid2rp.get_namespaces(qs)
        req.environ['openid2rp.ax'] = openid2rp.get_ax(qs, namespaces, signed)
        req.environ['openid2rp.sreg'] = openid2rp.get_sreg(qs, signed)
        return self.app(req.environ, start_response)

    def error(self, environ, start_response, msg):
        environ['openid2rp.error'] = msg
        return self.app(environ, start_response)



# Demo WSGI application using the openid2rp middleware.
import cgi

# This is the actual application. It displays a login box,
# and displays information about the user when login completes.
def application(environ, start_response):
    if 'openid2rp.error' in environ:
        start_response('401 Permission Denied', [('Content-type','text/plain')])
        return ['Something went wrong: '+environ['openid2rp.error']]
    if 'openid2rp.identifier' not in environ:
        # Display login box
        start_response('200 Ok', [('Content-type','text/html')])
        notice = environ.get('openid2rp.notice', '')
        if notice:
            notice = '<em>%s</em><br/>' % notice
        google = '/?openid_identifier=%s' % cgi.escape('https://www.google.com/accounts/o8/id')
        return ['<html><head><title>Login</title></head>'
                '<body>', notice, 'Login with OpenID:'
                '<form method="POST">'
                '<input name="openid_identifier" size="60"/>'
                '</form></body>',
                'Alternatively, log in directly with <a href="', google, '">Google</a>.',
                '</html>']
    # Display authentication results
    start_response('200 Ok', [('Content-type', 'text/html; charset=utf-8')])
    print environ['openid2rp.sreg'].items()
    return (['<html><head><title>Hello ',
             environ['openid2rp.identifier'].encode('utf-8'),
             '</title></head><body>'
             'We know the following information about you:<ul>']+
            ['<li>%s %s</li>' % (k, v) for k,v in environ['openid2rp.ax'].items()]+
            ['<li>%s %s</li>' % (k, v) for k,v in environ['openid2rp.sreg'].items()]+
            ['</ul></body></html>'])

# Wrap the application with the openid2rp middleware
# Web frameworks may offer to perform this wrapping by means of configuration,
# instead of code
from openid2rp.wsgi import Openid2Middleware
application = Openid2Middleware(application)

# Put this into a web container
# We use wsgiref here to keep the dependencies of this package low
from wsgiref.simple_server import make_server
server = make_server('', 6543, application)
print 'Running on http://localhost:6543/'
server.serve_forever()



from django.db import models
from django.contrib.auth.models import User
from django.utils.encoding import smart_unicode

class UserOpenID(models.Model):
	user = models.ForeignKey(User, related_name='openids')
	uri = models.CharField(max_length=255, blank=False, null=False)
	insert_date = models.DateTimeField(null=False, blank=False, auto_now_add=True, editable=False)
	last_modified = models.DateTimeField(null=False, blank=False, auto_now=True, editable=False)
	def __unicode__(self):
		return smart_unicode("OpenID claim for user '"+str(self.user)+"': "+str(self.uri))
		
class OpenIDSession(models.Model):
	claimedId = models.CharField(max_length=255, blank=False, null=False)
	assoc_handle = models.CharField(max_length=255, blank=False, null=False)
	mac_key = models.CharField(max_length=255, blank=False, null=False)
	expiration_date = models.DateTimeField(null=False, blank=False, editable=False)
	session_type = models.CharField(max_length=255, blank=True, null=True)
	assoc_type = models.CharField(max_length=255, blank=True, null=True)
	ns = models.CharField(max_length=255, blank=True, null=True)
	def __unicode__(self):
		return smart_unicode("OpenID session for '"+str(self.claimedId)+"', expires at "+str(self.expiration_date))
	
class OpenIDNonce(models.Model):
	nonce = models.CharField(max_length=255, blank=False, null=False)
	expiration_date = models.DateTimeField(null=False, blank=False, editable=False)
	def __unicode__(self):
		return smart_unicode("OpenID nonce '"+str(self.nonce)+"', expires at "+str(self.expiration_date))



# -*- coding: utf-8 -*-
# Django authentication backend, based on openid2rp
# Copyright Peter Tröger, 2010
# Licensed under the Academic Free License, version 3

from django.conf import settings
from django.contrib.auth.models import User
from openid2rp.django.models import UserOpenID, OpenIDSession, OpenIDNonce
from django.http import HttpResponse
from django.db.models import Q
from django.contrib.auth.models import AnonymousUser
from django.conf import settings
import openid2rp
import datetime

# we need to accept some difference between provider time and our time, for nonce and session checking
# Martin says: Use the Kerberos default
shiftVal = int(getattr(settings, 'OPENID2RP_MAXTIMESHIFT', '5'))
maxTimeShift=datetime.timedelta(minutes=shiftVal)
delayVal = int(getattr(settings, 'OPENID2RP_MAXLOGINDELAY', '5'))
maxLoginDelay=datetime.timedelta(minutes=delayVal)

# mirror the AX key name dict, so that the app does not need openid2rp at all
AX = openid2rp.AX

class IncorrectURIError(Exception):
	pass

class IncorrectClaimError(Exception):
	pass
	
class MissingSessionError(Exception):
	pass

class AuthenticationError(Exception):
	pass

class ReplayAttackError(Exception):
	pass

class IncompleteAnswerError(Exception):
	pass
	
class MultipleClaimUsageError(Exception):
	pass
	
class TookTooLongError(Exception):
	pass

def cleanup():
	try:
		# delete all expired nonces 
		entries=OpenIDNonce.objects.filter(Q(expiration_date__lt = datetime.datetime.utcnow()))
		for e in entries:
			e.delete()
	except:
		pass
	try:
		# delete all expired sessions
		entries=OpenIDSession.objects.filter(Q(expiration_date__lt = datetime.datetime.utcnow()))
		for e in entries:
			e.delete()
	except:
		pass
	
def storeNonce(nonce):
	global maxTimeShift, maxLoginDelay
	db = OpenIDNonce()
	db.nonce=nonce
	db.expiration_date = datetime.datetime.utcnow() + maxTimeShift + maxLoginDelay
	db.save()
	
def knownNonce(n):
	try:
		result = OpenIDNonce.objects.get(nonce=n)
	except:
		return False
	return True
	
def storeSession(session, claim):
	global maxTimeShift

	db = OpenIDSession()
	db.assoc_handle=session['assoc_handle']
	db.mac_key=session['mac_key']
	db.claimedId=claim
	# Expire session in provider-given amount of seconds, consider possible shift
	db.expiration_date = datetime.datetime.utcnow() + datetime.timedelta(seconds=long(session['expires_in'])) - maxTimeShift
	if 'assoc_type' in session:
		db.assoc_type=session['assoc_type']
	if 'ns' in session:
		db.ns=session['ns']
	if 'session_type' in session:
		db.session_type=session['session_type']
	db.save()

def getSessionByHandle(handle):
	try:
		result = OpenIDSession.objects.filter(Q(assoc_handle=handle)).values()[0]
	except:
		return None
	# we can live with the fact that the original expires_in field is missing, since openid2rp is not checking it anyway
	return result

def getSessionByClaim(claim):
	try:
		result = OpenIDSession.objects.filter(Q(claimedId=claim)).values()[0]
	except:
		return None
	# we can live with the fact that the original expires_in field is missing, since openid2rp is not checking it anyway
	return result

	
def linkOpenID(user, openid):
	claim=UserOpenID(user=user, uri=openid)
	claim.save()

def getOpenIDs(user):
	return UserOpenID.objects.filter(user=user).values_list('uri', flat=True)

def unlinkOpenID(user, openid):
	res = UserOpenID.objects.filter(Q(user=user, uri=openid))
	for r in res:
		r.delete()

def preAuthenticate(uri, answer_url, 
					sreg = (('nickname', 'email'), ()),
					ax = ((openid2rp.AX.email, openid2rp.AX.first, openid2rp.AX.last), ())):

	cleanup()
	try:
		kind, claimedId = openid2rp.normalize_uri(uri)			
	except Exception, e:
		raise IncorrectURIError(str(e))
	res = openid2rp.discover(claimedId)
	if res != None:
		services, url, op_local = res
		# re-use session in order to avoid provider roundtrip here
		session = getSessionByClaim(claimedId)
		if not session:
			session = openid2rp.associate(services, url)
			storeSession(session, claimedId)
		redirect_url=openid2rp.request_authentication( services, url, session['assoc_handle'], answer_url, claimedId, op_local, sreg=sreg, ax=ax )
		response=HttpResponse()
		response['Location']=redirect_url
		response.status_code=303
		return response
	else:
		raise IncorrectClaimError()

class Backend:	
	def get_user(self, user_id):
		try:
			return User.objects.get(pk=user_id)
		except User.DoesNotExist:
			return None

	def authenticate(self, **credentials):
		global maxTimeShift, maxLoginDelay

		# the default way for telling Django that this auth backend does not fit
		if not ("openidrequest" in credentials):
			raise TypeError

		request=credentials['openidrequest']
		query=request.META['QUERY_STRING']
		handle = request.GET['openid.assoc_handle']
		session=getSessionByHandle(handle)
		if not session:
			raise MissingSessionError
		try:
			signed=openid2rp.authenticate(session, query)
		except Exception, e:
			raise AuthenticationError(str(e))
		# check for replay attack
		nonce = request.GET['openid.response_nonce']
		timestamp = openid2rp.parse_nonce(nonce)		
		if timestamp < datetime.datetime.utcnow() - maxTimeShift - maxLoginDelay: 
			raise TookTooLongError()
		elif knownNonce(nonce):
			raise ReplayAttackError()
		storeNonce(nonce)

		# provider-based auth returns claim id, OpenID not (if I got that right) - in this case we take the one stored in the session found by assocHandle
		if 'openid.claimed_id' in request.GET:
			if 'claimed_id' not in signed:
				raise IncompleteAnswerError()
			claimedId = request.GET['openid.claimed_id']
		else:
			if 'identity' not in signed:
				raise IncompleteAnswerError()
			claimedId = session.claimedId

		# look up OpenID claim string in local database
		idrecord=UserOpenID.objects.filter(Q(uri=claimedId))
		if len(idrecord)>1:
			# more than one user has this claimID, which is definitly wrong
			raise MultipleClaimUsageError()
		elif len(idrecord)<1:
			# No user has this OpenID claim string assigned
			user = AnonymousUser()
		else:
			user=idrecord[0].user

		# inactive users are handled by the later login() method, so we can return them here too
		user.openid_claim = claimedId
		user.openid_ax = openid2rp.get_ax(query, openid2rp.get_namespaces(query), signed)
		user.openid_sreg = openid2rp.get_sreg(query, signed)
		user.openid_email = openid2rp.get_email(query)
		return user
		
